from django import template
from django.template import defaulttags
from django.template.defaulttags import URLNode


register = template.Library()
esi_tmpl = '<esi:include src="%s" />'

class EsiTemplateTagError(Exception):
    pass

class EsiNode(URLNode):
    def __init__(self, *args, **kwargs):
        super(EsiNode, self).__init__(*args, **kwargs)
        if '/' in self.view_name:
            # An actual URL has been passed instead of a view name.
            self.raw_url = self.view_name
            self.view_name = None
        else:
            self.raw_url = None

    def render(self, context):
        try:
            context['_esi']['used'] = True
        except KeyError:
            raise EsiTemplateTagError('The esi templatetag requires the esi context processor, but it isn\'t present.')

        url = self.raw_url or super(EsiNode, self).render(context)

        if self.asvar:
            url = context[self.asvar]
            context[self.asvar] = esi_tmpl % url
            return ''
        else:
            return esi_tmpl % url

@register.tag
def esi(parser, token):
    url_node = defaulttags.url(parser, token)
    return EsiNode(url_node.view_name, url_node.args, url_node.kwargs,
        url_node.asvar)



from django import template
from django.core.urlresolvers import reverse
from django.template import Token, Parser, TOKEN_BLOCK
import fudge
import random

from .._utils import TestCase

from ... import context_processors
from ...templatetags.esi import EsiNode
from ...templatetags.esi import esi

def create_context():
    request = fudge.Fake()
    context = template.Context()
    context.update(context_processors.esi(request))
    return context

def create_token(content):
    return Token(TOKEN_BLOCK, content)

class TestOfEsiNode(TestCase):
    def test_renders_actual_code(self):
        context = create_context()
        node = esi(Parser([]), create_token('esi hello_world'))
        result = node.render(context)

        expected_url = reverse('hello_world')
        self.assertEquals(result, '<esi:include src="%s" />' % expected_url)

    def test_renders_relative_esi(self):
        context = create_context()
        node = esi(Parser([]), create_token('esi ./blah/'))
        result = node.render(context)

        expected_url = './blah/'
        self.assertEquals(result, '<esi:include src="%s" />' % expected_url)

    def test_renders_kwargs(self):
        context = create_context()
        number = random.randint(100, 200)
        node = esi(Parser([]), create_token('esi hello_number number=%s' % number))
        result = node.render(context)

        expected_url = reverse('hello_number', kwargs={'number': number})
        self.assertEquals(result, '<esi:include src="%s" />' % expected_url)

    def test_sets_esi_used_to_true_on_context(self):
        context = create_context()
        node = EsiNode('hello_world', [], {}, None)
        node.render(context)

        self.assertTrue(context['_esi']['used'])

class TestOfEsiHandler(TestCase):
    def test_extracts_view_out_of_templatetag_call(self):
        random_view_name = 'hello_world_%d' % random.randint(100, 200)
        token = fudge.Fake()
        token.expects('split_contents').returns(('esi', random_view_name))
        fudge.clear_calls()

        result = esi(None, token)
        self.assertEquals(result.view_name, random_view_name)

        fudge.verify()

    def test_is_registered_as_a_templatetag_at_esi(self):
        library = template.get_library('esi')
        self.assert_('esi' in library.tags)
        self.assert_(library.tags['esi'] is esi)

    def test_can_be_rendered_from_a_template(self):
        raw_template = """
        {% load esi %}
        {% esi hello_world %}
        """

        t = template.Template(raw_template)
        context = create_context()
        result = t.render(context).strip()
        expected_url = reverse('hello_world')
        self.assertEquals(result, '<esi:include src="%s" />' % expected_url)




from django.http import HttpResponse, HttpResponseNotFound
from django.utils.cache import cc_delim_re, patch_vary_headers
from django.utils.http import http_date


def hello(request, number=None):
    if request.GET:
        pairs = ('%s = %s' % (key, request.GET[key]) for key in sorted(request.GET))
        return HttpResponse(', '.join(pairs))
    return HttpResponse(number or 'Hello World!')

def cookie_view(request, number):
    response = HttpResponse('C is for cookie.')
    response.set_cookie('a', 'apple')
    response.set_cookie('b', 'banana', path='/cookies/')
    response.set_cookie('number', number)
    return response

def last_modified(request, timestamp):
    response = HttpResponse(timestamp)
    response['Last-Modified'] = http_date(long(timestamp))
    return response

def vary(request):
    response = HttpResponse(request.GET['headers'])
    patch_vary_headers(response, cc_delim_re.split(request.GET['headers']))
    return response

def text(request):
    return HttpResponse('a' * 500)

def recursive_404(request):
    response = HttpResponseNotFound('<esi:include src="/recursive-404/" />')
    response._esi = {'used': True}
    return response



'''
This module began as a copy of django.test.client. Rather than import what we
need from there and create subclasses to implement the functionality we need,
we've copied over the file since django.test is unlikely to see the same
scrutiny or stability as modules intended for use in production code.
'''
import urllib
from urlparse import urlparse, urlunparse, urlsplit
import sys
import os
import re
import mimetypes
from Cookie import SimpleCookie
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from django.conf import settings
from django.core.handlers.base import BaseHandler
from django.core.handlers.wsgi import WSGIRequest
from django.core.signals import got_request_exception
from django.http import SimpleCookie, HttpRequest, QueryDict
from django.template import TemplateDoesNotExist
from django.utils.encoding import smart_str
from django.utils.http import urlencode
from django.utils.importlib import import_module
from django.utils.itercompat import is_iterable
from django.db import transaction, close_connection

BOUNDARY = 'BoUnDaRyStRiNg'
MULTIPART_CONTENT = 'multipart/form-data; boundary=%s' % BOUNDARY
CONTENT_TYPE_RE = re.compile('.*; charset=([\w\d-]+);?')

class FakePayload(object):
    """
    A wrapper around StringIO that restricts what can be read since data from
    the network can't be seeked and cannot be read outside of its content
    length. This makes sure that views can't do anything under the test client
    that wouldn't work in Real Life.
    """
    def __init__(self, content):
        self.__content = StringIO(content)
        self.__len = len(content)

    def read(self, num_bytes=None):
        if num_bytes is None:
            num_bytes = self.__len or 1
        assert self.__len >= num_bytes, "Cannot read more than the available bytes from the HTTP incoming data."
        content = self.__content.read(num_bytes)
        self.__len -= num_bytes
        return content

class LocalHandler(BaseHandler):
    """
    A HTTP Handler that can be used for fetching local URLs from
    application code.

    Based on django.test.client.ClientHandler, with the CSRF hacks removed.
    """
    def __call__(self, environ):
        from django.conf import settings

        # Set up middleware if needed. We couldn't do this earlier, because
        # settings weren't available.
        if self._request_middleware is None:
            self.load_middleware()

        request = WSGIRequest(environ)
        response = self.get_response(request)

        return response

def encode_multipart(boundary, data):
    """
    Encodes multipart POST data from a dictionary of form values.

    The key will be used as the form data name; the value will be transmitted
    as content. If the value is a file, the contents of the file will be sent
    as an application/octet-stream; otherwise, str(value) will be sent.
    """
    lines = []
    to_str = lambda s: smart_str(s, settings.DEFAULT_CHARSET)

    # Not by any means perfect, but good enough for our purposes.
    is_file = lambda thing: hasattr(thing, "read") and callable(thing.read)

    # Each bit of the multipart form data could be either a form value or a
    # file, or a *list* of form values and/or files. Remember that HTTP field
    # names can be duplicated!
    for (key, value) in data.items():
        if is_file(value):
            lines.extend(encode_file(boundary, key, value))
        elif not isinstance(value, basestring) and is_iterable(value):
            for item in value:
                if is_file(item):
                    lines.extend(encode_file(boundary, key, item))
                else:
                    lines.extend([
                        '--' + boundary,
                        'Content-Disposition: form-data; name="%s"' % to_str(key),
                        '',
                        to_str(item)
                    ])
        else:
            lines.extend([
                '--' + boundary,
                'Content-Disposition: form-data; name="%s"' % to_str(key),
                '',
                to_str(value)
            ])

    lines.extend([
        '--' + boundary + '--',
        '',
    ])
    return '\r\n'.join(lines)

def encode_file(boundary, key, file):
    to_str = lambda s: smart_str(s, settings.DEFAULT_CHARSET)
    content_type = mimetypes.guess_type(file.name)[0]
    if content_type is None:
        content_type = 'application/octet-stream'
    return [
        '--' + boundary,
        'Content-Disposition: form-data; name="%s"; filename="%s"' \
            % (to_str(key), to_str(os.path.basename(file.name))),
        'Content-Type: %s' % content_type,
        '',
        file.read()
    ]

class Client(object):
    """
    A class that can act as a client for testing purposes.

    It allows the user to compose GET and POST requests, and
    obtain the response that the server gave to those requests.
    The server Response objects are annotated with the details
    of the contexts and templates that were rendered during the
    process of serving the request.

    Client objects are stateful - they will retain cookie (and
    thus session) details for the lifetime of the Client instance.

    This is not intended as a replacement for Twill/Selenium or
    the like - it is here to allow testing against the
    contexts and templates produced by a view, rather than the
    HTML rendered to the end-user.
    """
    def __init__(self, cookies=None, handler_class=LocalHandler, **defaults):
        self.handler = handler_class()
        self.defaults = {'SERVER_NAME': 'localserver'}
        self.defaults.update(defaults)
        self.cookies = SimpleCookie(cookies or {})
        self.exc_info = None
        self.errors = StringIO()

    def _session(self):
        """
        Obtains the current session variables.
        """
        if 'django.contrib.sessions' in settings.INSTALLED_APPS:
            engine = import_module(settings.SESSION_ENGINE)
            cookie = self.cookies.get(settings.SESSION_COOKIE_NAME, None)
            if cookie:
                return engine.SessionStore(cookie.value)
        return {}
    session = property(_session)

    def request(self, **request):
        """
        The master request method. Composes the environment dictionary
        and passes to the handler, returning the result of the handler.
        Assumes defaults for the query environment, which can be overridden
        using the arguments to the request.
        """
        environ = {
            'HTTP_COOKIE':       self.cookies.output(header='', sep='; '),
            'PATH_INFO':         '/',
            'QUERY_STRING':      '',
            'REMOTE_ADDR':       '127.0.0.1',
            'REQUEST_METHOD':    'GET',
            'SCRIPT_NAME':       '',
            'SERVER_NAME':       'testserver',
            'SERVER_PORT':       '80',
            'SERVER_PROTOCOL':   'HTTP/1.1',
            'wsgi.version':      (1,0),
            'wsgi.url_scheme':   'http',
            'wsgi.errors':       self.errors,
            'wsgi.multiprocess': True,
            'wsgi.multithread':  False,
            'wsgi.run_once':     False,
        }
        environ.update(self.defaults)
        environ.update(request)

        try:
            response = self.handler(environ)
        except TemplateDoesNotExist, e:
            # If the view raises an exception, Django will attempt to show
            # the 500.html template. If that template is not available,
            # we should ignore the error in favor of re-raising the
            # underlying exception that caused the 500 error. Any other
            # template found to be missing during view error handling
            # should be reported as-is.
            if e.args != ('500.html',):
                raise

        # Update persistent cookie data.
        if response.cookies:
            self.cookies.update(response.cookies)

        return response


    def get(self, path, data={}, follow=False, **extra):
        """
        Requests a response from the server using GET.
        """
        parsed = urlparse(path)
        r = {
            'CONTENT_TYPE':    'text/html; charset=utf-8',
            'PATH_INFO':       urllib.unquote(parsed[2]),
            'QUERY_STRING':    urlencode(data, doseq=True) or parsed[4],
            'REQUEST_METHOD': 'GET',
            'wsgi.input':      FakePayload('')
        }
        r.update(extra)

        response = self.request(**r)
        if follow:
            response = self._handle_redirects(response, **extra)
        return response

    def post(self, path, data={}, content_type=MULTIPART_CONTENT,
             follow=False, **extra):
        """
        Requests a response from the server using POST.
        """
        if content_type is MULTIPART_CONTENT:
            post_data = encode_multipart(BOUNDARY, data)
        else:
            # Encode the content so that the byte representation is correct.
            match = CONTENT_TYPE_RE.match(content_type)
            if match:
                charset = match.group(1)
            else:
                charset = settings.DEFAULT_CHARSET
            post_data = smart_str(data, encoding=charset)

        parsed = urlparse(path)
        r = {
            'CONTENT_LENGTH': len(post_data),
            'CONTENT_TYPE':   content_type,
            'PATH_INFO':      urllib.unquote(parsed[2]),
            'QUERY_STRING':   parsed[4],
            'REQUEST_METHOD': 'POST',
            'wsgi.input':     FakePayload(post_data),
        }
        r.update(extra)

        response = self.request(**r)
        if follow:
            response = self._handle_redirects(response, **extra)
        return response

    def head(self, path, data={}, follow=False, **extra):
        """
        Request a response from the server using HEAD.
        """
        parsed = urlparse(path)
        r = {
            'CONTENT_TYPE':    'text/html; charset=utf-8',
            'PATH_INFO':       urllib.unquote(parsed[2]),
            'QUERY_STRING':    urlencode(data, doseq=True) or parsed[4],
            'REQUEST_METHOD': 'HEAD',
            'wsgi.input':      FakePayload('')
        }
        r.update(extra)

        response = self.request(**r)
        if follow:
            response = self._handle_redirects(response, **extra)
        return response

    def options(self, path, data={}, follow=False, **extra):
        """
        Request a response from the server using OPTIONS.
        """
        parsed = urlparse(path)
        r = {
            'PATH_INFO':       urllib.unquote(parsed[2]),
            'QUERY_STRING':    urlencode(data, doseq=True) or parsed[4],
            'REQUEST_METHOD': 'OPTIONS',
            'wsgi.input':      FakePayload('')
        }
        r.update(extra)

        response = self.request(**r)
        if follow:
            response = self._handle_redirects(response, **extra)
        return response

    def put(self, path, data={}, content_type=MULTIPART_CONTENT,
            follow=False, **extra):
        """
        Send a resource to the server using PUT.
        """
        if content_type is MULTIPART_CONTENT:
            post_data = encode_multipart(BOUNDARY, data)
        else:
            post_data = data

        # Make `data` into a querystring only if it's not already a string. If
        # it is a string, we'll assume that the caller has already encoded it.
        query_string = None
        if not isinstance(data, basestring):
            query_string = urlencode(data, doseq=True)

        parsed = urlparse(path)
        r = {
            'CONTENT_LENGTH': len(post_data),
            'CONTENT_TYPE':   content_type,
            'PATH_INFO':      urllib.unquote(parsed[2]),
            'QUERY_STRING':   query_string or parsed[4],
            'REQUEST_METHOD': 'PUT',
            'wsgi.input':     FakePayload(post_data),
        }
        r.update(extra)

        response = self.request(**r)
        if follow:
            response = self._handle_redirects(response, **extra)
        return response

    def delete(self, path, data={}, follow=False, **extra):
        """
        Send a DELETE request to the server.
        """
        parsed = urlparse(path)
        r = {
            'PATH_INFO':       urllib.unquote(parsed[2]),
            'QUERY_STRING':    urlencode(data, doseq=True) or parsed[4],
            'REQUEST_METHOD': 'DELETE',
            'wsgi.input':      FakePayload('')
        }
        r.update(extra)

        response = self.request(**r)
        if follow:
            response = self._handle_redirects(response, **extra)
        return response

    def _handle_redirects(self, response, **extra):
        "Follows any redirects by requesting responses from the server using GET."

        response.redirect_chain = []
        while response.status_code in (301, 302, 303, 307):
            url = response['Location']
            scheme, netloc, path, query, fragment = urlsplit(url)

            redirect_chain = response.redirect_chain
            redirect_chain.append((url, response.status_code))

            if scheme:
                extra['wsgi.url_scheme'] = scheme

            # The test client doesn't handle external links,
            # but since the situation is simulated in test_client,
            # we fake things here by ignoring the netloc portion of the
            # redirected URL.
            response = self.get(path, QueryDict(query), follow=False, **extra)
            response.redirect_chain = redirect_chain

            # Prevent loops
            if response.redirect_chain[-1] in response.redirect_chain[0:-1]:
                break
        return response



import random

from ._utils import TestCase
from .. import http_client


class ClientTest(TestCase):

    def test_plain_url(self):
        client = http_client.Client()
        self.assertEqual(client.get('/hello/').content, u'Hello World!')

    def test_kwargs_url(self):
        rand = random.randint(100, 200)
        client = http_client.Client()
        self.assertEqual(client.get('/hello/%d/' % rand).content, unicode(rand))

    def test_get_params(self):
        client = http_client.Client()
        expected = 'a = apple, b = banana'
        self.assertEqual(client.get('/hello/?b=banana&a=apple').content, expected)



import hashlib
import re

from django.core.urlresolvers import resolve
from django.core.cache import cache
from django.http import HttpResponse

from .utils import replace_esi_tags, gzip_response_content, \
    gunzip_response_content


esi_tag_re = re.compile(r'<esi:include src="(?P<url>[^"]+?)"\s*/>', re.I)

class IncludeEsiMiddleware(object):
    def process_response(self, request, response):
        esi_status = getattr(request, '_esi', {'used': False})
        if not esi_status['used']:
            return response

        # There is the possibility that GZipMiddleware has already been loaded by
        # the time we get this.  This is an uncommon case (and one advised against
        # in Django documentation), but when it happens we need to be able to work.
        #
        # Note: Running GZipMiddleware prior to the IncludeEsiMiddleware causes the
        # response to be compressed, decompressed, and then recompressed again.
        # Nine times out of ten, this is **not** what you want, but in the rare
        # instance that it is, this will continue to work as expected.
        is_gzipped = response.get('Content-Encoding', None) == 'gzip'
        if is_gzipped:
            gunzip_response_content(response)

        replace_esi_tags(request, response)

        if is_gzipped:
            gzip_response_content(request, response)

        return response

class EsiHeaderMiddleware(object):
    def process_response(self, request, response):
        if hasattr(request, '_esi'):
            response['X-ESI'] = 'true'
        return response



from django.conf.urls.defaults import patterns, url, handler404

urlpatterns = patterns('armstrong.esi.tests.esi_support.views',
    url(r'^hello/$', 'hello', name='hello_world'),
    url(r'^hello/(?P<number>\d+)/$', 'hello', name='hello_number'),
    url(r'^cookies/(?P<number>\d+)/$', 'cookie_view', name='cookie_view'),
    url(r'^last-modified/(?P<timestamp>\d+)/$', 'last_modified', name='last_modified'),
    url(r'^vary/$', 'vary', name='vary'),
    url(r'^500chars/$', 'text', name='text'),
    url(r'^recursive-404/$', 'recursive_404', name='recursive_404'),
)



from django.http import HttpRequest
import random

import fudge
from ._utils import TestCase
from ._utils import with_fake_request

from ..context_processors import esi

class TestOfEsiContextProcessor(TestCase):
    @with_fake_request
    def test_adds_esi_token_to_context(self, request):
        result = esi(request)
        self.assert_('_esi' in result, msg='sanity check')
        self.assertFalse(result['_esi']['used'])

        request._esi['used'] = True
        result = esi(request)
        self.assert_('_esi' in result, msg='sanity check')
        self.assert_(result['_esi']['used'])



from cStringIO import StringIO
from email.utils import parsedate
import gzip
import logging
import re
import time
from urlparse import urljoin

from django.conf import settings
from django.http import HttpResponse
from django.middleware.gzip import GZipMiddleware
from django.utils.cache import cc_delim_re
from django.utils.datastructures import MultiValueDict
from django.utils.http import http_date

from . import http_client

try:
    from logging import NullHandler
except ImportError:
    # Compatibility mode for 2.6.x
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


log = logging.getLogger('armstrong.esi')
log.addHandler(NullHandler())
esi_tag_re = re.compile(r'<esi:include src="(?P<url>[^"]+?)"\s*/>', re.I)

def reduce_vary_headers(response, additional):
    '''Merges the Vary header values so all headers are included.'''
    original = response.get('Vary', None)
    if original is not None:
        additional.append(original)
    # Keep track of normalized, lowercase header names in seen_headers while
    # maintaining the order and case of the header names in final_headers.
    seen_headers = set()
    final_headers = []
    for vary_value in additional:
        headers = cc_delim_re.split(vary_value)
        for header in headers:
            if header.lower() in seen_headers:
                continue
            seen_headers.add(header.lower())
            final_headers.append(header)
    response['Vary'] = ', '.join(final_headers)

def reduce_last_modified_headers(response, additional):
    '''Sets Last-Modified to the latest of all of the header values.'''
    dates = additional
    if 'Last-Modified' in response:
        dates.append(response['Last-Modified'])
    dates = [time.mktime(parsedate(date_str)) for date_str in dates]
    latest = max(dates)
    response['Last-Modified'] = http_date(latest)

HEADERS_TO_MERGE = {
    'Vary': reduce_vary_headers,
    'Last-Modified': reduce_last_modified_headers,
}

def merge_fragment_headers(response, fragment_headers):
    '''
    Given fragment_headers, a MultiValueDict or other mapping of header names to
    values, add the header values to the response as appropriate.
    '''
    for header, reduce_func in HEADERS_TO_MERGE.items():
        if not header in fragment_headers:
            continue
        if hasattr(fragment_headers, 'getlist'):
            values = fragment_headers.getlist(header)
        else:
            value = fragment_headers.get(header)
            values = [value] if value is not None else []
        reduce_func(response, values)

def merge_fragment_cookies(response, fragment_cookies):
    '''
    Merges the fragment and response cookies.

    Set the fragment cookies in the order they occurred, then set the main
    response cookie last.
    '''
    if not fragment_cookies:
        return
    cookies = fragment_cookies[0]
    cookies_to_reduce = fragment_cookies[1:]
    cookies_to_reduce.append(response.cookies)
    for cookie_obj in cookies_to_reduce:
        for key, morsel in cookie_obj.items():
            # To set Morsels as cookie values directly, we need to bypass
            # BaseCookie.__setitem__.
            dict.__setitem__(cookies, key, morsel)
    response.cookies = cookies

def gunzip_response_content(response):
    '''
    If the response has already been compressed, gunzip it so we can modify
    the text.
    '''
    f = gzip.GzipFile(fileobj=StringIO(response.content))
    response.content = f.read()
    f.close()
    del response['Content-Encoding']

def gzip_response_content(request, response):
    GZipMiddleware().process_response(request, response)

def build_full_fragment_url(request, url):
    if url.startswith('/'):
        return url
    else:
        return urljoin(request.path, url)

# TODO: Test this independently of the middleware
# TODO: Reduce the lines of codes and varying functionality of this code so its
#       tests can be reduced in complexity.
def replace_esi_tags(request, response):
    process_errors = getattr(settings, 'ESI_PROCESS_ERRORS', False)
    fragment_headers = MultiValueDict()
    fragment_cookies = []
    request_data = {
        'cookies': request.COOKIES,
        'HTTP_REFERER': request.build_absolute_uri(),
        'HTTP_X_ESI_FRAGMENT': True,
    }

    replacement_offset = 0
    for match in esi_tag_re.finditer(response.content):
        url = build_full_fragment_url(request, match.group('url'))

        if response.status_code == 200 or process_errors:
            client = http_client.Client(**request_data)
            fragment = client.get(url)
        else:
            fragment = HttpResponse()

        if fragment.status_code != 200:
            # Remove the error content so it isn't added to the page.
            fragment.content = ''
            extra = {'data': {
                'fragment': fragment.__dict__,
                'request': request.__dict__,
            }}
            log.error('ESI fragment %s returned status code %s' %
                (url, fragment.status_code), extra=extra)

        start = match.start() + replacement_offset
        end = match.end() + replacement_offset
        response.content = '%s%s%s' % (response.content[:start],
            fragment.content, response.content[end:])
        replacement_offset += len(fragment.content) - len(match.group(0))

        for header in HEADERS_TO_MERGE:
            if header in fragment:
                fragment_headers.appendlist(header, fragment[header])
        if fragment.cookies:
            fragment_cookies.append(fragment.cookies)

    merge_fragment_headers(response, fragment_headers)
    merge_fragment_cookies(response, fragment_cookies)



from django.conf import settings
from django.http import HttpRequest
from django.http import HttpResponse
from django.middleware.gzip import GZipMiddleware
from django.utils.cache import cc_delim_re
from django.utils.http import http_date
import fudge
import hashlib
import itertools
import math
import random
import re
import urllib

from ._utils import TestCase
from ._utils import with_fake_request
from .esi_support.views import recursive_404

from .. import middleware
from ..middleware import IncludeEsiMiddleware, EsiHeaderMiddleware
from ..utils import gunzip_response_content


MIDDLEWARES = [
    'armstrong.esi.middleware.IncludeEsiMiddleware',
    'armstrong.esi.middleware.EsiHeaderMiddleware',
]

def full_process_response(request, response, gzip=False):
    if gzip:
        response = GZipMiddleware().process_response(request, response)
    response = EsiHeaderMiddleware().process_response(request, response)
    response = IncludeEsiMiddleware().process_response(request, response)
    return response

def patch_settings(new_settings):
    patches = []
    added_settings = []
    for key, value in new_settings.items():
        if hasattr(settings, key):
            patches.append(fudge.patcher.patch_object(settings, key, value))
        else:
            setattr(settings, key, value)
            added_settings.append(key)
    return patches, added_settings

def restore_settings(patches, added_settings):
    for patch in patches:
        patch.restore()
    for setting in added_settings:
        delattr(settings, setting)

class TestMiddleware(TestCase):

    @with_fake_request
    def test_returns_unmodified_response_on_non_esi_response(self, request):
        original_content = str(random.randint(1000, 2000))
        response = HttpResponse(original_content)
        new_response = full_process_response(request, response)
        self.assertEqual(original_content, new_response.content)

    @with_fake_request
    def test_replaces_esi_tags_with_actual_response(self, request):
        rand = random.randint(100, 200)
        url = '/hello/%d/' % rand

        request.provides('get_full_path').returns('/')
        request.provides('build_absolute_uri').returns('http://example.com/')
        request.has_attr(_esi={'used': True})

        response = HttpResponse()
        esi_tag = '<esi:include src="%s" />' % url
        response.content = esi_tag

        patch_data = patch_settings({'MIDDLEWARE_CLASSES': MIDDLEWARES})

        result = full_process_response(request, response)
        self.assertFalse(re.search(esi_tag, result.content), msg='sanity check')
        self.assertEquals(result.content, str(rand))

        restore_settings(*patch_data)

    @with_fake_request
    def test_replaces_esi_tags_in_gzipped_response(self, request):
        request.provides('get_full_path').returns('/')
        request.provides('build_absolute_uri').returns('http://example.com/')
        request.has_attr(_esi={'used': True})
        request.has_attr(META={'HTTP_ACCEPT_ENCODING': 'gzip'})

        response = HttpResponse()
        esi_tag = '<esi:include src="/500chars/" />'
        response.content = '%s%s%s' % ('z' * 250, esi_tag, 'z' * 250)

        patch_data = patch_settings({'MIDDLEWARE_CLASSES': MIDDLEWARES})

        result = full_process_response(request, response, gzip=True)
        self.assertFalse(re.search(esi_tag, result.content), msg='sanity check')
        self.assertTrue(result.get('Content-Encoding', None) == 'gzip')

        gunzip_response_content(result)
        self.assertTrue(re.search('a' * 500, result.content))

        restore_settings(*patch_data)

    @with_fake_request
    def test_replaces_relative_url_esi(self, request):
        rand = random.randint(100, 200)
        url = './%d/' % rand

        request.has_attr(path='/hello/')
        request.provides('build_absolute_uri').returns('http://example.com/')
        request.has_attr(_esi={'used': True})

        response = HttpResponse()
        esi_tag = '<esi:include src="%s" />' % url
        response.content = esi_tag
        fudge.clear_calls()

        patch_data = patch_settings({'MIDDLEWARE_CLASSES': MIDDLEWARES})

        result = full_process_response(request, response)
        self.assertFalse(re.search(esi_tag, result.content), msg='sanity check')
        self.assertEquals(result.content, str(rand))

        restore_settings(*patch_data)

    @with_fake_request
    def check_content_permutation(self, request, permutation, chunk_results):
        request_url = '/page-with-esi-tags/'
        request.has_attr(_esi={'used': True})
        request.provides('get_full_path').returns(request_url)
        request.provides('build_absolute_uri').returns(
            'http://example.com%s' % request_url)

        patch_data = patch_settings({'MIDDLEWARE_CLASSES': MIDDLEWARES})

        response = HttpResponse(''.join(permutation))
        expected_result = ''.join(chunk_results[chunk] for chunk in permutation)
        result = full_process_response(request, response)
        self.assertEqual(result.content, expected_result)

        restore_settings(*patch_data)

    # TODO: This is a slow test that should be adjusted for so it doesn't take up
    #       > 80% of the execution time.
    def test_replaces_multiple_esi_tags_correctly(self):
        static_text_options = (
            '',  # Allow ESI tags to end up next to each other
            'abcdefg',
            'abcdefghijklmnopqrstuvwxyz' * 10,
        )
        main_page_chunks = dict(zip(static_text_options, static_text_options))

        # Make sure some of the includes generate content longer than the ESI
        # tag itself and some shorter.
        number_lengths = (1, 10, 50, 100)
        for length in number_lengths:
            num = random.randint(math.pow(10, length-1), math.pow(10, length) - 1)
            content = '<esi:include src="/hello/%s/" />' % num
            main_page_chunks[content] = str(num)

        for permutation in itertools.permutations(main_page_chunks, 4):
            self.check_content_permutation(permutation, main_page_chunks)

    def get_cookie_test_objs(self, request):
        number = random.randint(100, 200)
        fragment_url = '/cookies/%d/' % number
        request_url = '/page-with-esi-tag/'
        response = HttpResponse('<esi:include src="%s" />' % fragment_url)
        request.has_attr(_esi={'used': True})
        request.provides('get_full_path').returns(request_url)
        request.provides('build_absolute_uri').returns(
            'http://example.com%s' % request_url)
        return request, response, number

    @with_fake_request
    def test_merges_cookies(self, request):
        request, response, number = self.get_cookie_test_objs(request)
        result = full_process_response(request, response)

        self.assertEqual(result.cookies['a'].value, 'apple')
        self.assertEqual(result.cookies['b'].value, 'banana')
        self.assertEqual(result.cookies['number'].value, str(number))
        self.assertEqual(result.cookies['b']['path'], '/cookies/')

    @with_fake_request
    def test_main_response_cookies_take_precedence(self, request):
        request, response, number = self.get_cookie_test_objs(request)
        response.set_cookie('a', 'alligator')
        result = full_process_response(request, response)

        self.assertEqual(result.cookies['a'].value, 'alligator')

    @with_fake_request
    def check_if_max_time_is_last_modified(self, request, main_response_time,
      fragment_time_1, fragment_time_2, max_time):
        fragment_urls = ['/last-modified/%s/' % num for num in
            (fragment_time_1, fragment_time_2)]
        html = '\n'.join('<esi:include src="%s" />' % url for url in
            fragment_urls)
        response = HttpResponse(html)
        response['Last-Modified'] = http_date(main_response_time)

        request_url = '/page-with-esi-tags/'
        request.has_attr(_esi={'used': True})
        request.provides('get_full_path').returns(request_url)
        request.provides('build_absolute_uri').returns(
            'http://example.com%s' % request_url)

        response = full_process_response(request, response)
        self.assertTrue(response['Last-Modified'], max_time)

    def test_merges_last_modified(self):
        time_a = random.randint(10000, 20000)
        time_b = random.randint(40000, 50000)
        time_c = random.randint(60000, 70000)

        max_time = http_date(time_c)
        self.check_if_max_time_is_last_modified(time_a, time_b, time_c, max_time)
        self.check_if_max_time_is_last_modified(time_b, time_c, time_a, max_time)
        self.check_if_max_time_is_last_modified(time_c, time_a, time_b, max_time)

    @with_fake_request
    def check_merged_vary_header(self, request, main_header, fragment_header_1,
      fragment_header_2):
        fragment_urls = ['/vary/?headers=%s' % urllib.quote_plus(vary) for vary in
            (fragment_header_1, fragment_header_2) if vary]
        html = '\n'.join('<esi:include src="%s" />' % url for url in
            fragment_urls)
        response = HttpResponse(html)
        if main_header:
            response['Vary'] = main_header

        request_url = '/page-with-esi-tags/'
        request.has_attr(_esi={'used': True})
        request.provides('get_full_path').returns(request_url)
        request.provides('build_absolute_uri').returns(
            'http://example.com%s' % request_url)

        result = full_process_response(request, response)
        vary_result = result.get('Vary', '')

        for header in (main_header, fragment_header_1, fragment_header_2):
            if not header:
                continue
            vary_fields = cc_delim_re.split(header)
            for field in vary_fields:
                self.assertEqual(len(re.findall(field, vary_result, re.I)), 1)
        
    def test_merges_vary(self):
        vary_sets = (
            (None, None, 'Cookie'),
            (None, 'cookie', 'Cookie'),
            (None, 'Accept-Encoding', 'Cookie'),
            (None, 'Accept-Encoding', 'Cookie, Accept-Language'),
            (None, 'Accept-Encoding, Accept-Language', 'Cookie'),
            ('Accept-Encoding', 'Cookie', 'Cookie'),
        )

        for headers in vary_sets:
            self.check_merged_vary_header(*headers)

    @with_fake_request
    def test_recursive_error_pages(self, request):

        request.provides('get_full_path').returns('/')
        request.provides('build_absolute_uri').returns('http://example.com/')
        request.has_attr(_esi={'used': True})
        response = recursive_404(request)

        patch_data = patch_settings({
            'MIDDLEWARE_CLASSES': MIDDLEWARES,
            'ESI_PROCESS_ERRORS': False,
        })

        result = full_process_response(request, response)
        self.assertEqual(result.content, '')

        restore_settings(*patch_data)



from django.conf import settings
from django.http import HttpRequest
from django.test import TestCase as DjangoTestCase
import fudge

def with_fake_request(func):
    def inner(self, *args, **kwargs):
        request = fudge.Fake(HttpRequest)
        request.has_attr(COOKIES={})
        result = func(self, request, *args, **kwargs)

        fudge.verify()
        fudge.clear_expectations()
        return result
    return inner

class TestCase(DjangoTestCase):
    urls = 'armstrong.esi.tests.esi_support.urls'

    def setUp(self):
        self._original_settings = settings
        fudge.clear_expectations()
        fudge.clear_calls()

    def tearDown(self):
        settings = self._original_settings

